/**
	Brightness Control v1.0
	Copyright (C) 2012, Carlos DF
	c4rl0s.pt@gmail.com
	
	main.c: Brightness Control main code

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
#include <pspctrl.h>
#include <pspsdk.h>
#include <pspkernel.h>
#include <pspsysmem_kernel.h>
#include <pspdisplay_kernel.h>
#include <pspimpose_driver.h>
#include <string.h>
#include <psprtc.h>

#include "sysconhk.h"
#include "blit.h"
#include "minIni.h"

#define KERNEL_MODE 0x1000
#define HIGH_PRIORITY_THREAD 0x18
#define DEFAUT_PRIORITY_THREAD 0x12

#define ONE_SECOND 1000000

#define BRIGHTNESS_UP 1
#define BRIGHTNESS_DOWN 0

#define DISPLAY_MAX_TICK 200 	//Value for a timer used to display the brightness
#define DEFAUT_DISPLAY_FGCOLOR 0xFFFFFF  //Foregorund color is pure white
#define DEFAUT_DISPLAY_BGCOLOR 0x000000  //Background color is pure black

#define BRIGHTNESS_MESSAGE " BIGHTNESS %i "
#define BRIGHTNESS_CONFIG_FILE "brightness.bin"
#define INI_CONFIG_FILE "brightness.ini"

#define DEFAUT_KEY_COMBINATION (SYSCON_CTRL_LTRG + SYSCON_CTRL_LTRG)
#define DEFAUT_KEY_BUTTON_UP    SYSCON_CTRL_ALLOW_UP
#define DEFAUT_KEY_BUTTON_DN    SYSCON_CTRL_ALLOW_DN
#define DEFAUT_KEY_BUTTON_RT    SYSCON_CTRL_ALLOW_RT
#define DEFAUT_KEY_BUTTON_LT    SYSCON_CTRL_ALLOW_LT

PSP_MODULE_INFO("Brightness", KERNEL_MODE, 1, 1);
PSP_MAIN_THREAD_ATTR(0);

typedef struct _ConfigINI
{
		int ButtonKey;
		int ButtonUP;
		int ButtonDN;
		int ButtonRT;
		int ButtonLT;
		int LockControls;
		int DisplayMSG;
		int DisplayFGCOLOR;
		int DisplayBGCOLOR;
		char DisplayString[128];
} ConfigINI;

typedef struct _Config
{
		int brightness;
} Config;

static int running = 0;
static SceCtrlData raw_ctrl = {0};
static int displayTick = DISPLAY_MAX_TICK;
static u64 sleepTickLast = 0;
static u32 prevButtons, newButtons;
static int lastSavedBrightness = -1;

static ConfigINI configINI;
static Config config;
static char configPath[128];
static char iniPath[128];

void LoadINI(void)
{
	configINI.ButtonKey = ini_getlhex("brightness", "ButtonKey", DEFAUT_KEY_COMBINATION, iniPath);
	configINI.ButtonUP = ini_getlhex("brightness", "ButtonUP", DEFAUT_KEY_BUTTON_UP, iniPath);
	configINI.ButtonDN = ini_getlhex("brightness", "ButtonDN", DEFAUT_KEY_BUTTON_DN, iniPath);
	configINI.ButtonRT = ini_getlhex("brightness", "ButtonRT", DEFAUT_KEY_BUTTON_RT, iniPath);
	configINI.ButtonLT = ini_getlhex("brightness", "ButtonLT", DEFAUT_KEY_BUTTON_LT, iniPath);
	configINI.LockControls = ini_getlhex("brightness", "LockControls", 0, iniPath);
	configINI.DisplayMSG = ini_getlhex("brightness", "DisplayMSG", 1, iniPath);
	configINI.DisplayFGCOLOR = ini_getlhex("brightness", "DisplayFGCOLOR", DEFAUT_DISPLAY_FGCOLOR, iniPath);
	configINI.DisplayBGCOLOR = ini_getlhex("brightness", "DisplayBGCOLOR", DEFAUT_DISPLAY_BGCOLOR, iniPath);
	ini_gets("brightness", "DisplayString", BRIGHTNESS_MESSAGE, configINI.DisplayString, 128, iniPath);
}

int GetConfig(Config *cfg)
{
	SceUID fd;
	u32 k1;
   
	k1 = pspSdkSetK1(0);
	memset(cfg, 0, sizeof(*cfg));
	fd = sceIoOpen(configPath, PSP_O_RDONLY, 0644);

	if (fd < 0) {
		pspSdkSetK1(k1);
		return -1;
	}

	if (sceIoRead(fd, cfg, sizeof(*cfg)) != sizeof(*cfg)) {
		sceIoClose(fd);
		pspSdkSetK1(k1);
		return -2;
	}

	sceIoClose(fd);
	pspSdkSetK1(k1);

	return 0;
}

int SetConfig(Config *cfg)
{
	u32 k1;
	SceUID fd;

	k1 = pspSdkSetK1(0);
	sceIoRemove(configPath);
	fd = sceIoOpen(configPath, PSP_O_WRONLY | PSP_O_CREAT | PSP_O_TRUNC, 0777);

	if (fd < 0) {
		pspSdkSetK1(k1);
		return -1;
	}

	if (sceIoWrite(fd, cfg, sizeof(*cfg)) != sizeof(*cfg)) {
		sceIoClose(fd);
		pspSdkSetK1(k1);
		return -1;
	}

	sceIoClose(fd);
	pspSdkSetK1(k1);

	return 0;
}

void SetConfigPath(const char *argp)
{
	strcpy(configPath, argp);
	strrchr(configPath, '/')[1] = 0;
	strcpy(configPath + strlen(configPath), BRIGHTNESS_CONFIG_FILE);
	//
	strcpy(iniPath, argp);
	strrchr(iniPath, '/')[1] = 0;
	strcpy(iniPath + strlen(iniPath), INI_CONFIG_FILE);
}

int getBrightness(){
	int ret;
	u32 k1;
    k1 = pspSdkSetK1(0);
	sceDisplayGetBrightness(&ret, 0);
    pspSdkSetK1(k1);
    return ret;
}
 
void setBrightness(int value){
	u32 k1;
    k1 = pspSdkSetK1(0);
	config.brightness = value;
	// -Set the Backlight level (0-4)
	// 0 = 60
	// 1 = 72
	// 2 = 84
	// 3 = 44
	// 4 = 0 (screen off)
	if (value < 44) {
		sceImposeSetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS, 3);
	} else if (value < 60) {
		sceImposeSetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS, 0);
	} else if (value < 72) {
		sceImposeSetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS, 1);
	} else if (value < 84) {
		sceImposeSetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS, 2);
	} else {
		sceImposeSetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS, 3);
	}	
	sceDisplaySetBrightness(value, 0);
    pspSdkSetK1(k1);
}

/*
void setBrightnessFade(int value){
	// Get the brightness
	int currBrightness = getBrightness();
	// Make a fade efect..
	int loop = 1;
	int brightValue = currBrightness;
	while (brightValue != value) {
		if (value > currBrightness) {
			brightValue = currBrightness + loop;
		} else {
			brightValue = currBrightness - loop;
		}
		setBrightness(brightValue);
		sceKernelDelayThread(ONE_SECOND / 40);
		loop++;
	}
} */

int main_thread(SceSize args, void *argp)
{
	// Load the INI file
	LoadINI();
	// Load brightness settings
	if (GetConfig(&config) != 0) {
		config.brightness = getBrightness();
	} else {
		 // Check for min and max brightness values
		if (config.brightness < 11) config.brightness = 11;
		if (config.brightness > 100) config.brightness = 100;
		// Set the brightness
		setBrightness(config.brightness);
	}
	lastSavedBrightness = config.brightness;
	// Install system hook
	install_syscon_hook();
	sceKernelDelayThread(ONE_SECOND / 2);
	// Show brightness message on start up
	displayTick = 0;
	//
	while (running)  
	{

		sceKernelDelayThreadCB(ONE_SECOND / 4);
	
		while (displayTick < DISPLAY_MAX_TICK) {
		
			if (configINI.DisplayMSG) {
				char msg[128];
				sprintf(msg, configINI.DisplayString, getBrightness());
				blit_string(27, 30, msg, configINI.DisplayFGCOLOR, configINI.DisplayBGCOLOR);
				sceDisplayWaitVblankStart();
			} else {
				sceKernelDelayThreadCB(100);
			}
			
			displayTick++;
		}
		
		// Save configs if need
		if (lastSavedBrightness != config.brightness) {
			SetConfig(&config);
			lastSavedBrightness = config.brightness;
		}
		
	}
	
	sceKernelExitDeleteThread(0);
	return 0;
}

void changeBrightness(int step, int direction) {
	
	u64 sleepNowTick;
	int sleepInterval;
	
	sceRtcGetCurrentTick(&sleepNowTick);

	if (step == 1) {
		sleepInterval = 100000; // 0.10 secunds
	} else {
		sleepInterval = 300000; // 0.30 secunds
	}
	// Make a simple function frezz..
	if (sleepTickLast > 0) {
		if ((sleepNowTick - sleepTickLast) < sleepInterval) {
			return;		
		}
	}

	sceRtcGetCurrentTick(&sleepTickLast);
	
	int brightness = getBrightness();
	
	if (direction == BRIGHTNESS_UP) {
		brightness = brightness + step;
	} else if (direction == BRIGHTNESS_DOWN) {
		brightness = brightness - step;
	}

	if (step != 1) {
		if (brightness < 11) brightness = 11;
	}
	if (brightness > 100) brightness = 100;
 
	/*
	if (step == 1) { */
		setBrightness(brightness);
	/*} else { 
		setBrightnessFade(brightness);
	}*/
	
	// Show brightness message
	displayTick = 0;
}

int module_start(SceSize args, void *argp)
{
	running = 1;
	SetConfigPath(argp);
	/* Create a thread */
	int thid = sceKernelCreateThread("brightness_thread", main_thread, HIGH_PRIORITY_THREAD, 0x1000, 0, NULL); 
	if (thid > 0) {
		sceKernelStartThread(thid, args, argp);	
	}
	return 0;
}

int module_stop(SceSize args, void *argp)
{
	running = 0;
	uninstall_syscon_hook();
	return 0;
}

void syscon_ctrl(sceSysconPacket *packet)
{
	newButtons = 0;
 
	switch(packet->rx_response)
	{
	case 0x08:

	case 0x07:
		newButtons = syscon_get_dword(packet->rx_data);
		newButtons = ~newButtons;

		prevButtons = raw_ctrl.Buttons;
		raw_ctrl.Buttons = newButtons;

		if (newButtons & configINI.ButtonKey) {

			if (newButtons & configINI.ButtonUP) {
			
				changeBrightness(1,  BRIGHTNESS_UP);
				if (configINI.LockControls) {
					newButtons &= ~configINI.ButtonKey;
					newButtons &= ~configINI.ButtonUP;
				}
				
			} else if (newButtons & configINI.ButtonDN) {
			
				changeBrightness(1,  BRIGHTNESS_DOWN);
				if (configINI.LockControls) {
					newButtons &= ~configINI.ButtonKey;
					newButtons &= ~configINI.ButtonDN;
				}
				
			} else if (newButtons & configINI.ButtonRT) {
			
				changeBrightness(10, BRIGHTNESS_UP);
				if (configINI.LockControls) {
					newButtons &= ~configINI.ButtonKey;
					newButtons &= ~configINI.ButtonRT;
				}
				
			} else if (newButtons & configINI.ButtonLT) {
			
				changeBrightness(10, BRIGHTNESS_DOWN);
				if (configINI.LockControls) {
					newButtons &= ~configINI.ButtonKey;
					newButtons &= ~configINI.ButtonLT;
				}
				
			}
		}

		if (newButtons & SYSCON_CTRL_LCD) {
			// If the screen is turned off and back now to ON
			// set the right brightness
			// if (!(newButtons & SYSCON_DEV_ON)) {
				// u32 k1 = pspSdkSetK1(0);
				// int ret = sceImposeGetParam(PSP_IMPOSE_BACKLIGHT_BRIGHTNESS);
				// if (ret > 3) {
					// setBrightness(lastSavedBrightness);
				// }
				// pspSdkSetK1(k1);
				// if (!(newButtons & SYSCON_DEV_ON)) {
					// int brightness = getBrightness();
					// if (brightness < 11) {
						// setBrightness(lastSavedBrightness);
						// newButtons &= ~SYSCON_CTRL_LCD;
					// }
				// }
			// }
			//			
			displayTick = 0;
		}
		
	}
	
	syscon_put_dword(packet->rx_data,~newButtons);
	syscon_make_checksum(&packet->rx_sts);
 
}

/* Exported function returns the address of module_info */
void* getModuleInfo(void)
{
	return (void *) &module_info;
}
